from pyramid.settings import asbool

BOOL_SETTINGS = [
    "broker_connection_retry",
    "broker_use_ssl",
    "elasticsearch_retry_on_timeout",
    "elasticsearch_save_meta_as_text",
    "enable_utc",
    "redis_backend_use_ssl",
    "redis_retry_on_timeout",
    "redis_socket_keepalive",
    "result_backend_always_retry",
    "result_extended",
    "result_persistent",
    "task_acks_late",
    "task_acks_on_failure_or_timeout",
    "task_always_eager",
    "task_create_missing_queues",
    "task_default_rate_limit",
    "task_eager_propagates",
    "task_ignore_result",
    "task_inherit_parent_priority",
    "task_publish_retry",
    "task_reject_on_worker_lost",
    "task_remote_tracebacks",
    "task_send_sent_event",
    "task_store_errors_even_if_ignored",
    "task_track_started",
    "worker_direct",
    "worker_disable_rate_limits",
    "worker_enable_remote_control",
    "worker_hijack_root_logger",
    "worker_log_color",
    "worker_pool_restarts",
    "worker_redirect_stdouts",
    "worker_send_task_events",
]


INT_SETTINGS = [
    "beat_max_loop_interval",
    "beat_sync_every",
    "broker_connection_max_retries",
    "broker_pool_limit",
    "broker_transport_options.max_retries",
    "broker_transport_options.visibility_timeout",
    "cassandra_port",
    "elasticsearch_max_retries",
    "redis_max_connections",
    "result_backend_base_sleep_between_retries_ms",
    "result_backend_max_retries",
    "result_backend_max_sleep_between_retries_ms",
    "result_cache_max",
    "result_expires",
    "task_default_priority",
    "task_protocol",
    "task_publish_retry_policy.max_retries",
    "task_queue_max_priority",
    "worker_concurrency",
    "worker_max_memory_per_child",
    "worker_max_tasks_per_child",
    "worker_prefetch_multiplier",
]


FLOAT_SETTINGS = [
    "azureblockblob_retry_increment_base",
    "azureblockblob_retry_initial_backoff_sec",
    "azureblockblob_retry_max_attempts",
    "broker_connection_timeout",
    "broker_heartbeat",
    "broker_heartbeat_checkrate",
    "control_queue_expires",
    "control_queue_ttl",
    "elasticsearch_timeout",
    "event_queue_expires",
    "event_queue_ttl",
    "redis_socket_connect_timeout",
    "redis_socket_timeout",
    "result_chord_join_timeout",
    "result_chord_retry_interval",
    "task_publish_retry_policy.interval_max",
    "task_publish_retry_policy.interval_start",
    "task_publish_retry_policy.interval_step",
    "task_soft_time_limit",
    "task_time_limit",
    "worker_lost_wait",
    "worker_proc_alive_timeout",
    "worker_timer_precision",
]


LIST_SETTINGS = [
    "accept_content",
    "cassandra_servers",
    "imports",
    "include",
    "result_accept_content",
]


def extract_celery_settings(settings):
    celery_settings = dict()
    filtered = (
        (k[7:], v) for (k, v) in settings.items() if k.startswith("celery.")
    )
    for key, value in filtered:
        *parts, final_part = key.split(".")
        location = celery_settings
        for part in parts:
            location = location.setdefault(part, dict())
        if key in BOOL_SETTINGS:
            value = asbool(value)
        elif key in INT_SETTINGS:
            value = int(value)
        elif key in FLOAT_SETTINGS:
            value = float(value)
        elif key in LIST_SETTINGS:
            value = [x.strip() for x in value.split(",")]
        location[final_part] = value
    return celery_settings



import celery
import venusian
from pyramid.interfaces import (
    PHASE1_CONFIG,
    PHASE2_CONFIG,
    PHASE3_CONFIG,
    IRequestFactory,
)
from pyramid.request import Request
from pyramid.scripting import prepare

from .events import BeforeDeferTask, CeleryWorkerProcessInit
from .settings import extract_celery_settings
from .taskderivers import apply_task_derivers

# If set, pyramid_tasks will use this Celery application rather than make a new
# one.  This is necessary because when running a worker via an ini file (see
# celery.py), we need to create the app immediately but have no way of passing
# in the app when loading the app.
global_app = None


def set_global_app(app):
    global global_app
    global_app = app


def includeme(config):
    settings = config.get_settings()
    if global_app is None:
        app = celery.Celery(
            autofinalize=False,
            set_as_current=False,
        )
    else:
        app = global_app
    _connect_worker_process_init(config.registry)
    app.conf.update(extract_celery_settings(settings))
    app.pyramid_config = config
    config.registry["pyramid_tasks.app"] = app
    config.registry["pyramid_tasks.task_map"] = dict()
    config.add_directive("make_celery_app", make_celery_app)
    config.add_directive("register_task", register_task)
    config.add_directive(
        "add_periodic_task", add_periodic_task, action_wrap=True
    )
    config.action(
        ("celery", "finalize"),
        _get_app(config.registry).finalize,
        order=PHASE2_CONFIG,
    )
    config.add_request_method(defer_task)
    config.add_request_method(defer_task_with_options)
    config.add_request_method(defer_task, "delay_task")  # Legacy
    config.add_request_method(get_task_result)
    config.add_request_method(current_task, property=True)
    config.include(".taskderivers")


def _get_app(registry):
    return registry["pyramid_tasks.app"]


def make_celery_app(config):
    config.commit()
    return _get_app(config.registry)


def register_task(config, func, name=None, **kwargs):
    """
    Register a new task with Celery.

    """
    registry = config.registry
    app = _get_app(registry)
    name = name or app.gen_task_name(func.__name__, func.__module__)

    def register():
        derived = apply_task_derivers(config, func, name, kwargs)
        handler = _make_task_handler(config.registry, derived)
        task = app.task(
            handler,
            name=name,
            shared=False,
            bind=True,
            **kwargs,
        )
        config.registry["pyramid_tasks.task_map"][func] = task

    discriminator = ("register task", name)
    config.action(discriminator, register, order=PHASE1_CONFIG)


def _make_task_handler(registry, func):
    """
    Return a task handler that creates a new request and injects it as the
    first positional argument in the wrapped function.

    """

    def handler(self, *args, **kwargs):
        headers = self.request.headers
        environ = headers.get("environ") if headers else None
        request = _make_request(registry, environ)
        with prepare(request=request, registry=registry):
            return func(request, *args, **kwargs)

    return handler


def _make_request(registry, environ):
    request_factory = registry.queryUtility(IRequestFactory, default=Request)
    request = request_factory.blank("/", environ=environ)
    return request


def task(**kwargs):
    """
    Decorator to register a new task.

    """

    def wrapper(func):
        def callback(scanner, name, wrapped):
            scanner.config.register_task(wrapped, **kwargs)

        venusian.attach(func, callback, category="pyramid")
        return func

    return wrapper


def _get_task(registry, func_or_name):
    """
    Get task by function or name.

    """
    task_map = registry["pyramid_tasks.task_map"]
    if isinstance(func_or_name, str):
        return _get_app(registry).tasks[func_or_name]
    elif func_or_name in task_map:
        return task_map[func_or_name]
    else:
        raise ValueError("Not a valid task.")


def defer_task(request, func_or_name, *args, **kwargs):
    """
    Add a task to the work queue.  All positional and keyword arguments passed
    into this method will be passed through to the task.  Convenience function
    for ``defer_task_with_options``.

    """
    return defer_task_with_options(
        request,
        func_or_name,
        args=args,
        kwargs=kwargs,
    )


def defer_task_with_options(
    request, func_or_name, args=tuple(), kwargs=None, **options
):
    """
    Add a task to the work queue.  ``func_or_name`` can either by the name of
    the task (str) or the task function itself.  ``args`` and ``kwargs`` will
    be passed into the task.  Additional options will be passed into
    ``Task.apply_async``.  Please refer to Celery documentation to see options.

    """
    kwargs = kwargs if kwargs is not None else dict()
    task = _get_task(request.registry, func_or_name)
    request.registry.notify(
        BeforeDeferTask(
            request=request,
            task=task,
            args=args,
            kwargs=kwargs,
            options=options,
        )
    )
    return task.apply_async(args, kwargs, **options)


def add_periodic_task(
    config, schedule, func_or_name, args=(), kwargs=None, **opts
):
    """
    Add a period task.

    """

    def add():
        task = _get_task(config.registry, func_or_name)
        _get_app(config.registry).add_periodic_task(
            schedule,
            task,
            args,
            kwargs,
            **opts,
        )

    config.action(None, add, order=PHASE3_CONFIG)


def get_task_result(request, task_id):
    """
    Get a result object from celery.

    """
    return _get_app(request.registry).AsyncResult(task_id)


def current_task(request):
    """
    Return the task currently being executed.

    """
    return _get_app(request.registry).current_worker_task


def _connect_worker_process_init(registry):
    """
    Connect a signal handler to Celery's ``worker_process_init`` module.

    """

    def handler(sender=None, **kwargs):
        registry.notify(CeleryWorkerProcessInit(registry))

    signal = celery.signals.worker_process_init
    signal.connect(handler, weak=False)



from celery import Celery, signals
from pyramid.paster import get_app

from . import set_global_app

celery = Celery(autofinalize=False)


def add_preload_arguments(parser):
    parser.add_argument(
        "-i", "--ini", default=None, help="Paste ini configuration file."
    )
    parser.add_argument(
        "--ini-var",
        default=None,
        help="Comma separated list of key=value to pass to ini.",
    )


celery.user_options["preload"].add(add_preload_arguments)


@signals.user_preload_options.connect
def on_preload_parsed(options, **kwargs):
    """
    Load Pyramid application.

    """
    ini_location = options["ini"]
    ini_vars = options["ini_var"]
    if ini_location is None:
        print("You must provide the --ini argument")  # noqa: T201
        exit(-1)

    options = {}
    if ini_vars is not None:
        for pairs in ini_vars.split(","):
            key, value = pairs.split("=")
            options[key] = value

    set_global_app(celery)
    # We don't do anything with this, just want to trigger configuration.
    get_app(ini_location, options=options)



from zope.interface import Attribute, Interface


class ITaskDeriver(Interface):
    def __call__(task, info):
        """
        Derive a new task from the supplied task.

        View options, package information and registry are available on
        ``info``, an instance of
        :class:`pyramid_tasks.interfaces.ITaskDeriverInfo`.

        The ``task`` is a callable accepting ``(request, *args, **kwargs)``.

        """


class ITaskDeriverInfo(Interface):
    """
    An object implementing this interface is passed to every task deriver
    during configuration.

    """

    registry = Attribute(
        'The "current" application registry where the task was created'
    )
    package = Attribute(
        "The current package where the configuration statement was found"
    )
    name = Attribute("The name of the task.")
    options = Attribute("The options passed to the task.")
    original_func = Attribute(
        "The original function being transformed into a task."
    )


class ITaskDerivers(Interface):
    """Interface for task derivers list."""

    def add(self, name, factory, after=None, before=None):
        """
        Add a task deriver to the list.

        """

    def sorted(self):
        """
        Return a sorted list of task derivers.

        """



from pyramid_tasks.settings import extract_celery_settings


def test_extract_settings_basic():
    settings = {
        "celery.foo": "bar",
        "celery.fizz": "buzz",
    }
    assert extract_celery_settings(settings) == {
        "foo": "bar",
        "fizz": "buzz",
    }


def test_extract_settings_bool():
    settings = {
        "celery.task_acks_late": "yes",
        "celery.task_ignore_result": False,
    }
    assert extract_celery_settings(settings) == {
        "task_acks_late": True,
        "task_ignore_result": False,
    }


def test_extract_settings_int():
    settings = {
        "celery.worker_concurrency": "3",
    }
    assert extract_celery_settings(settings) == {
        "worker_concurrency": 3,
    }


def test_extract_settings_float():
    settings = {
        "celery.task_soft_time_limit": "1.5",
        "celery.task_time_limit": "3",
    }
    assert extract_celery_settings(settings) == {
        "task_soft_time_limit": 1.5,
        "task_time_limit": 3.0,
    }


def test_extract_settings_list():
    settings = {
        "celery.result_accept_content": "json, pickle",
    }
    assert extract_celery_settings(settings) == {
        "result_accept_content": ["json", "pickle"],
    }


def test_extract_settings_dict():
    settings = {
        "celery.broker_transport_options.max_retries": "3",  # int coercion
        "celery.broker_transport_options.foo.fizz": "buzz",
        "celery.broker_transport_options.bar": "baz",
    }
    assert extract_celery_settings(settings) == {
        "broker_transport_options": {
            "max_retries": 3,
            "bar": "baz",
            "foo": {
                "fizz": "buzz",
            },
        },
    }



from contextlib import contextmanager
from unittest.mock import Mock

import pytest
from celery.contrib.testing.worker import start_worker
from pyramid.scripting import prepare
from pyramid.testing import testConfig as _testConfig

from pyramid_tasks import BeforeDeferTask


@pytest.fixture
def test_config():
    settings = {
        "celery.worker_hijack_root_logger": False,
        "celery.worker_log_color": False,
        "celery.accept_content": "json",
        "celery.enable_utc": True,
        "celery.timezone": "UTC",
        "celery.broker_url": "memory://",
        "celery.result_backend": "cache+memory://",
        "celery.broker_heartbeat": 0,
    }
    with _testConfig(settings=settings, autocommit=False) as config:
        config.include("pyramid_tasks")
        yield config


@contextmanager
def make_request_with_worker(config):
    config.commit()
    with prepare(registry=config.registry) as env:
        with make_worker(config):
            yield env["request"]


@contextmanager
def make_worker(config):
    with start_worker(config.make_celery_app(), perform_ping_check=False):
        yield None


def test_simple_integration(test_config):
    def add_task(request, x, y):
        return x + y

    test_config.register_task(add_task, name="add")
    with make_request_with_worker(test_config) as request:
        assert request.defer_task("add", 2, 3).get() == 5


def test_options_integration(test_config):
    """
    Test `defer_task_with_options`.

    """

    def test_task(request):
        from celery import current_task

        return current_task.request.foo

    test_config.register_task(test_task, name="test")
    with make_request_with_worker(test_config) as request:
        result = request.defer_task_with_options(
            "test", headers={"foo": "bar"}
        ).get()
    assert result == "bar"


def test_legacy_integration(test_config):
    """
    Test legacy `delay_task` call.

    """

    def add_task(request, x, y):
        return x + y

    test_config.register_task(add_task, name="add")
    with make_request_with_worker(test_config) as request:
        result = request.delay_task("add", 2, 3).get()
    assert result == 5


def test_task_by_function_integration(test_config):
    def add_task(request, x, y):
        return x + y

    test_config.register_task(add_task, name="add")
    with make_request_with_worker(test_config) as request:
        result = request.defer_task(add_task, 2, 3).get()
    assert result == 5


def test_task_venusian_integration(test_config):
    test_config.include("tests.pkgs.venusianapp")
    with make_request_with_worker(test_config) as request:
        result = request.defer_task("venusian_add_task", 2, 3).get()
    assert result == 5


def test_get_task_result_integration(test_config):
    def add_task(request, x, y):
        return x + y

    test_config.register_task(add_task, name="add")
    with make_request_with_worker(test_config) as request:
        task_id = request.defer_task("add", 2, 3).id
        result = request.get_task_result(task_id)
    assert result.get() == 5


def test_task_request_argument_integration(test_config):
    task = Mock(return_value=None)
    test_config.register_task(task, name="test")
    app = test_config.make_celery_app()
    with make_worker(test_config):
        app.tasks["test"].apply_async().get()
    request = task.call_args[0][0]
    assert request.registry == test_config.registry


def test_task_deriver_integration(test_config):
    test_config.include("tests.pkgs.deriverapp")

    db = test_config.registry["db"]
    with db:
        db["counter"] = 3

    with make_request_with_worker(test_config) as request:
        assert request.defer_task("increment", 4).get() == 7


def test_task_environ_integration(test_config):
    def add_task(request):
        return request.environ["foo"]

    test_config.register_task(add_task, name="add")
    environ = {"foo": "bar"}
    with make_request_with_worker(test_config) as request:
        result = request.defer_task_with_options(
            "add",
            headers={"environ": environ},
        ).get()
    assert result == "bar"


def test_transaction_integration(test_config):
    test_config.include("tests.pkgs.transactionapp")

    with make_request_with_worker(test_config) as request:
        request.defer_task("increment", 1).get()
        request.defer_task("increment", 3).get()
        assert request.defer_task("total").get() == 4


def test_task_before_apply_integration(test_config):
    def add_headers(event):
        assert event.request
        assert event.task
        assert event.args == tuple()
        assert event.kwargs == dict()
        event.options.setdefault("headers", dict())["foo"] = "bar"

    def task(request):
        return request.current_task.request.foo

    test_config.add_subscriber(add_headers, BeforeDeferTask)
    test_config.register_task(task)
    with make_request_with_worker(test_config) as request:
        assert request.defer_task(task).get() == "bar"



class BeforeDeferTask:
    """
    This event is emitted before a task is queued (``Task.apply_async()``).

    The ``kwargs`` attribute is a dictionary containing the keyword arguments
    to be passed into ``Task.apply_async``.  Modifying this dictionary will
    modify the keywords arguments passed.

    """

    def __init__(self, request, task, args, kwargs, options):
        self.request = request
        """The current request."""
        self.task = task
        """The task being queued."""
        self.task_name = task.name
        """The name of the task being queued."""
        self.args = args
        """The position arguments the task will be invoked with."""
        self.kwargs = kwargs
        """The keyword arguments the task will be invoked with."""
        self.options = options
        """The additional options being passed in to ``Task.apply_async``."""


class CeleryWorkerProcessInit:
    """
    This is event is emitted Celery's ``worker_process_init`` signal.  Useful
    for disposing of any multiprocess unsafe resources.

    See more:  https://docs.celeryq.dev/en/stable/userguide/signals.html

    """

    def __init__(self, registry):
        self.registry = registry
        """The current Pyramid registry."""



"""
For backwards compatibility.  For new proejcts, use
``pyramid_tasks.contrib.pyramid_tm`` instead.

"""

from .contrib.pyramid_tm import transaction_task_deriver


def includeme(config):
    config.include(".contrib.pyramid_tm")


__all__ = [
    "transaction_task_deriver",
]



"""
Emulates ``pyramid_tm`` functionality inside of Celery.

"""


def includeme(config):
    config.add_task_deriver(transaction_task_deriver)


def transaction_task_deriver(task, info):
    def deriver(request, *args, **kwargs):
        environ = request.environ
        if environ.get("tm.active"):
            return task(request, *args, **kwargs)

        manager = request.tm
        environ["tm.active"] = True
        environ["tm.manager"] = manager

        with manager:
            return task(request, *args, **kwargs)

    return deriver



"""
Task derivers are the task analog of Pyramid's view derivers.

"""

from pyramid.interfaces import PHASE0_CONFIG
from pyramid.util import TopologicalSorter

from .interfaces import ITaskDerivers


def includeme(config):
    config.registry.registerUtility(TopologicalSorter(), ITaskDerivers)
    config.add_directive(
        "add_task_deriver", add_task_deriver, action_wrap=True
    )


def add_task_deriver(config, deriver, name=None, under=None, over=None):
    """
    Add a task deriver to the list.

    """
    deriver = config.maybe_dotted(deriver)

    if name is None:
        name = deriver.__name__

    discriminator = ("task deriver", name)
    intr = config.introspectable("task derivers", name, name, "task deriver")
    intr["name"] = name
    intr["deriver"] = deriver
    intr["under"] = under
    intr["over"] = over

    derivers = config.registry.getUtility(ITaskDerivers)

    def add():
        derivers.add(name, deriver, after=under, before=over)

    config.action(discriminator, add, order=PHASE0_CONFIG)


def apply_task_derivers(config, task, name, options):
    """
    Apply the task derivers to a task.

    """
    derivers = config.registry.getUtility(ITaskDerivers).sorted()
    info = TaskDeriverInfo(
        registry=config.registry,
        package=config.package,
        name=name,
        options=options,
        original_func=task,
    )
    for name, deriver in reversed(derivers):
        task = deriver(task, info)
    return task


class TaskDeriverInfo:
    def __init__(self, registry, package, name, options, original_func):
        self.registry = registry
        self.package = package
        self.name = name
        self.options = options
        self.original_func = original_func
